#error "The version indicated in ___VERSION is not supported by gambit.h"
#error "You are probably compiling C code generated by a Gambit compiler earlier or later than v4.9.3"
#error "___VERSION will be set to 409003 to reduce other errors"
#undef ___VERSION
#define ___VERSION 409003
#include "gambit.h"
